unit rkResTrans;

{$mode objfpc}{$H+}

{
  Main translation unit of package ResTrans,
  Reinhard Kalinke Nov 2021

  License Modified LGPL --- and no warranty whatsoever
}

{
Make sure to include at least LCLStrConsts.xy.po and, if relevant, your project.xy.po
in Project Options -> Resources (where 'xy' is the resp. language id).

Just adding LCLStrConsts.xy.po to your project's resources and rkDefTrans unit
to your uses is enough for a localized version of LCL strings.

If you want to include some FPC resources as well, just add the relevant po files
(cf https://forum.lazarus.freepascal.org/index.php/topic,56305.msg423740.html)
to your project's resources, they will be automatically searched for as well
(cf caDefaultUnits).

If there are other units to translate include the *.xy.po _and_ provide a list
of these additional unit names in project resources as well as a simple text
file with a line for each unit, eg ResTrans.txt (default value, cf csDefaultList).
So if you provide all necessary po-Files and a list of additional units which
transfers to a resource named 'RESTRANS' the standard translation will translate
to the default language of the target machine without any additional parameters.

Just let the IDE take care of the naming of resources so that for example
project.xy.po becomes 'PROJECT.XY', LCLStrConsts.xy.po -> 'LCLSTRCONSTS.XY'
and ResTrans.txt -> 'RESTRANS'.

In case of a custom list name or custom language id call TranslateFromResources
like this:

TranslateFromResources('xy','TransList',True); // force update of existing forms

If you only need standard translation just include rkDefTrans in your uses
(like unit DefaultTranslator supplied with Lazarus).

These routines won't throw an error if a particular resource is not included,
but of course the resp. target will not be translated either...
}

interface

uses Classes, SysUtils;

procedure TranslateFromResources(ForceUpdate: Boolean=True);
procedure TranslateFromResources(ALangID: string; ForceUpdate: Boolean=True);
procedure TranslateFromResources(ALangID, AUnitList: string; ForceUpdate: Boolean=True);

procedure TranslateUnitFromResource(const AUnitName: string; ALangID: string='');

implementation

uses Forms, LCLType, LResources, LCLTranslator, Translations, LazUTF8;

const
  csLCLStrConst  = 'lclstrconsts';
  csFpcRtlConst  = 'rtlconsts';
  csFpcSysConst  = 'sysconst';
  csFpcDBConst   = 'dbconst';

  caDefaultUnits: array of string = (
                                    csLCLStrConst,
                                    csFpcRtlConst,
                                    csFpcSysConst,
                                    csFpcDBConst
                                    );
  csDefaultList  = 'restrans';


procedure TranslateFromResources(ForceUpdate: Boolean);
begin
  TranslateFromResources('','',ForceUpdate);
end;

procedure TranslateFromResources(ALangID: string; ForceUpdate: Boolean);
begin
  TranslateFromResources(ALangID,'',ForceUpdate);
end;

procedure TranslateFromResources(ALangID, AUnitList: string; ForceUpdate: Boolean);
var LocalTranslator: TUpdateTranslator;
    Stream: TResourceStream;
    StrList: TStringList;
    ResName: string;
    POFile: TPOFile;
    hNd: THandle;
    i: integer;
begin

  LocalTranslator := nil;

  // ShortLanguageID is sufficient in most cases, if you need more or want
  // to process start parameters provide a more sophisticated method here
  if (ALangID = '') then
    LazGetShortLanguageID(ALangID);

  // first look for application po-file
  ResName := UpperCase(ChangeFileExt(ExtractFileName(ParamStrUTF8(0)),'.'+ALangID));
  hNd := FindResource(hInstance,PChar(ResName),RT_RCDATA);
  if (hNd <> 0) then
  begin
    Stream := TResourceStream.Create(hInstance,ResName,RT_RCDATA);
    try
      POFile := TPOFile.Create(Stream);
      TranslateResourceStrings(POFile);
      // TPOTranslator takes care of this TPOFile - so no freeing it
      LocalTranslator := TPOTranslator.Create(POFile);
    finally
      Stream.Free;
    end;
  end;

  // now translate DefaultUnits like LCLStrConsts etc...
  for i:= Low(caDefaultUnits) to High(caDefaultUnits) do
    TranslateUnitFromResource(caDefaultUnits[i],ALangID);

  // if no custom list is supplied try default list
  if (AUnitList = '') then
    AUnitList := csDefaultList;

  // finally work through the list if present
  ResName := UpperCase(AUnitList);
  hNd := FindResource(hInstance,PChar(ResName),RT_RCDATA);
  if (hNd <> 0) then
  begin
    Stream := TResourceStream.Create(hInstance,ResName,RT_RCDATA);
    try
      StrList := TStringList.Create;
      try
        StrList.LoadFromStream(Stream);
        for i:=0 to pred(StrList.Count) do
          if (StrList.Strings[i] > '') then
            TranslateUnitFromResource(StrList.Strings[i],ALangID);
      finally
        StrList.Free;
      end;
    finally
      Stream.Free;
    end;
  end;

  if Assigned(LocalTranslator) then
  begin
    // assign the new standard translator
    if Assigned(LRSTranslator) then
      LRSTranslator.Free;
    LRSTranslator := LocalTranslator;

    // as stated in SetDefaultLang() this might not be safe
    // for calls from initialization section or .lpr!
    // (tested on Linux - it is... and one more check:)
    if ForceUpdate and Assigned(Screen) then
    begin
      for i:=0 to pred(Screen.CustomFormCount) do
        LocalTranslator.UpdateTranslation(Screen.CustomForms[i]);
      for i:=0 to pred(Screen.DataModuleCount) do
        LocalTranslator.UpdateTranslation(Screen.DataModules[i]);
    end;
  end;
end;

procedure TranslateUnitFromResource(const AUnitName: string; ALangID: string='');
var Stream: TResourceStream;
    ResName: string;
    POFile: TPOFile;
    hNd: THandle;
begin
  if (ALangID = '') then
    LazGetShortLanguageID(ALangID);
  ResName := UpperCase(AUnitName+'.'+ALangID);
  hNd := FindResource(hInstance,PChar(ResName),RT_RCDATA);
  if (hNd <> 0) then
  begin
    Stream := TResourceStream.Create(hInstance,ResName,RT_RCDATA);
    try
      POFile := TPOFile.Create(Stream);
      try
        TranslateUnitResourceStrings(AUnitName,POFile);
      finally
        POFile.Free;
      end;
    finally
      Stream.Free;
    end;
  end;
end;

end.

